# @author Augustin Mortier
# @desc A-Profiles - Code for creating map json file to be used in V-Profiles

import json
import warnings
from pathlib import Path

import numpy as np
import xarray as xr


def make_map(path, yyyy: str, mm: str, mapname) -> None:
    # one map, per day, which collects the maximum extinction with no low-level clouds (<6km) at each station
    with open(Path(path, yyyy, mm, mapname), "w") as json_file:
        json.dump({}, json_file)


def add_to_map(fn, path, yyyy: str, mm: str, dd: str, mapname) -> None:
    # map collects the maximum extinction value with no low-level clouds (<6km) at each station at a hourly resolution
    # for each station, write an array with extinction values, and array with scenes for each hour of the day

    # read data
    vars_to_read = [
        "extinction",
        "retrieval_scene",
        "cloud_amount",
        "lidar_ratio",
        "aer_type",
        "mec",
    ]

    # in some cases, the file might be corrupted. Just skip it then.
    try:
        ds = xr.open_dataset(fn, engine="netcdf4", chunks=-1)[vars_to_read].load()
    except OSError:
        print(f"File {fn} is corrupted. Skipping...")
        return

    # calculate the max extinction and determine the scene for each hour of the day
    # need to convert time (from int to datetime) in order to use the resample method from xarray
    ds["time"] = ds.time.data.astype("datetime64[ms]").astype("datetime64[ns]")

    # in order to prevent some monotony issue, sort by time
    ds = ds.sortby("time")

    layers = ["0-6", "0-2", "2-4", "4-6"]
    max_ext = {}
    for layer in layers:
        zmin = float(layer.split("-")[0]) * 1000
        zmax = float(layer.split("-")[1]) * 1000
        zmin_asl = zmin + ds.attrs["station_altitude_t0"]
        zmax_asl = zmax + ds.attrs["station_altitude_t0"]

        # extinction for each hour
        max_ext_profiles = (
            ds.where((ds.altitude >= zmin_asl) & (ds.altitude < zmax_asl), drop=True)
            .extinction.resample(time="1h")
            .max()
            .data
        )

        # take the maximum value in each profile
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            max_ext[layer] = np.nanmax(max_ext_profiles, axis=1)

    # scene for each hour
    max_retrieval_scene = ds.retrieval_scene.resample(time="1h").max().data
    max_cloud_amount = ds.cloud_amount.resample(time="1h").max().data
    mean_lidar_ratio = ds.lidar_ratio.resample(time="1h").mean().data

    # make mec dictionary
    mec = dict()
    for i, aer_type in enumerate(ds.aer_type.data):
        mec[aer_type] = round(float(ds.mec[i].data), 3)

    # open current map
    with open(Path(path, yyyy, mm, mapname), "r") as json_file:
        data = json.load(json_file)
    json_file.close()

    # add new data to calendar data
    station_id = f'{ds.attrs["wigos_station_id"]}-{ds.attrs["instrument_id"]}'
    if station_id not in data:
        data[station_id] = {}
    data[station_id][dd] = {
        "attrs": {
            "coords": {
                "latitude": ds.attrs["station_latitude_t0"],
                "longitude": ds.attrs["station_longitude_t0"],
                "altitude": ds.attrs["station_altitude_t0"],
            },
            "l0_wavelength": ds.attrs["l0_wavelength"],
            "station_id": station_id,
            "station_name": ds.attrs["site_location"],
            "instrument_type": ds.attrs["instrument_type"],
        },
        "max_ext:0-6km": [
            round(ext, 4) if not np.isnan(ext) else None for ext in max_ext["0-6"]
        ],
        "max_ext:0-2km": [
            round(ext, 4) if not np.isnan(ext) else None for ext in max_ext["0-2"]
        ],
        "max_ext:2-4km": [
            round(ext, 4) if not np.isnan(ext) else None for ext in max_ext["2-4"]
        ],
        "max_ext:4-6km": [
            round(ext, 4) if not np.isnan(ext) else None for ext in max_ext["4-6"]
        ],
        "retrieval_scene": [
            retrieval_scene if not np.isnan(retrieval_scene) else None
            for retrieval_scene in max_retrieval_scene.tolist()
        ],
        "cloud_amount": [
            cloud_amount if not np.isnan(cloud_amount) else None
            for cloud_amount in max_cloud_amount.tolist()
        ],
        "lidar_ratio": [
            lidar_ratio if not np.isnan(lidar_ratio) else None
            for lidar_ratio in mean_lidar_ratio.tolist()
        ],
        "mec": mec,
    }

    # write new map
    with open(Path(path, yyyy, mm, mapname), "w") as json_file:
        json.dump(data, json_file)



# @author Augustin Mortier
# @desc A-Profiles - Code for creating climatology json file to be used in V-Profiles

from datetime import datetime
from pathlib import Path
import os

import numpy as np
import orjson
import xarray as xr


def compute_climatology(
    path, station_id, season_variables, all_variables, aerosols_only
):
    # get all files
    station_files = []
    for root, dirs, files in os.walk(path, followlinks=True):
        for file in files:
            if station_id in file and file.endswith(".nc"):
                station_files.append(os.path.join(root, file))

    station_files = sorted(station_files)
    try:
        # open dataset with xarray
        vars = (
            season_variables
            + all_variables
            + ["retrieval_scene", "cloud_amount", "scene"]
        )
        try:
            ds = xr.open_mfdataset(
                station_files,
                parallel=False,
                decode_times=True,
                chunks=-1,
                concat_dim="time",
                join="outer",
                data_vars=vars,
                coords="minimal",
                combine="nested",
                compat="override",
                drop_variables=["mec"],
            )[vars].load()
        except Exception as e:
            raise (e)

        ds = ds.sortby("time")
        ds = ds.drop_duplicates("time", keep="last")

        # store attributes which are destroyed by the resampling method
        attrs = ds.attrs
        # replace np.int by int
        for attr in attrs:
            if isinstance(attrs[attr], np.uint32):
                attrs[attr] = int(attrs[attr])

        # keep only clear scenes
        if aerosols_only:
            ds = ds.where((ds.retrieval_scene <= 1) & (ds.cloud_amount == 0))

        # daily resampling
        Dds = ds.resample(time="D").mean().compute()
        Dds["nprofiles"] = ds.scene.resample(time="D").count().compute()

        # define path
        clim_daily_path = Path(path, "climato_daily")

        # create directory if does not exist
        clim_daily_path.mkdir(parents=True, exist_ok=True)

        # write daily file
        Dds.to_netcdf(Path(clim_daily_path, f"AP_{station_id}_clim.nc"))

        # seasonal resampling
        Qds = ds.resample(time="QE").mean().compute()

        # add some statistics
        attrs["ndays"] = {
            "ndays": len(station_files),
            "since": str(ds.time.data[0]).split("T")[0],
        }
        attrs["today"] = datetime.today().strftime("%Y-%m-%d")

        # add number of days per season as a new variable
        Qds["ndays"] = ds.scene.resample(time="D").count().resample(time="QE").count()

        # add Z to time
        ds.coords["time"] = (ds["time"].astype(int) * 1e-6).astype(int)
        Qds.coords["time"] = (Qds["time"].astype(int) * 1e-6).astype(int)

        # round altitude to 3 decimals
        ds.coords["altitude"] = ds["altitude"].round(3)
        Qds.coords["altitude"] = Qds["altitude"].round(3)

        # select variables
        multivars_dict = {}

        # add seasonal variables
        for var in season_variables + ["ndays"]:
            multivars_dict[var] = Qds[var].round(6).to_dict()

        # add all variables
        for var in all_variables:
            multivars_dict[var] = ds[var].round(6).to_dict()

        # add attributes as separate key
        multivars_dict["attrs"] = attrs

        # define path
        clim_path = Path(path, "climato")
        # create directory if does not exist
        clim_path.mkdir(parents=True, exist_ok=True)

        # write data to json file
        with open(Path(clim_path, f"AP_{station_id}_clim.json"), "wb") as json_file:
            json_file.write(
                orjson.dumps(multivars_dict, option=orjson.OPT_SERIALIZE_NUMPY)
            )

    except Exception as e:
        print(f"Error encountered with {station_id}: {e}")



# @author Augustin Mortier
# @desc A-Profiles - Standard workflow

import aprofiles as apro
from pathlib import Path
import json
import warnings


def workflow(path, instruments_types, base_dir, CFG, verbose=False):
    apro_reader = apro.reader.ReadProfiles(path)
    profiles = apro_reader.read()

    nans = int(profiles.data.attenuated_backscatter_0.isnull().sum())
    total = profiles.data.attenuated_backscatter_0.size
    if (
        int(profiles.data.attenuated_backscatter_0.isnull().sum())
        / profiles.data.attenuated_backscatter_0.size
        >= 0.25
    ):
        warnings.warn(
            f"Error with {path}. attenuated_backscatter_0 variable has {nans}/{total} values as NaNs."
        )
        return

    # do the rest if instrument_type in the list
    if profiles.data.attrs["instrument_type"] in instruments_types:

        # extrapolation lowest layers
        profiles.extrapolate_below(z=260.0, inplace=True)

        # cloud detection
        if "cloud" in CFG["parameters"][profiles.data.instrument_type]:
            if "method" in CFG["parameters"][profiles.data.instrument_type]["cloud"]:
                cloud_method = CFG["parameters"][profiles.data.instrument_type][
                    "cloud"
                ]["method"]
            profiles.clouds(
                cloud_method,
                time_avg=1,
                zmin=300.0,
                thr_noise=5.0,
                thr_clouds=4.0,
                verbose=verbose,
            )

        # foc and pbl
        profiles.foc(zmin_cloud=300.0)
        profiles.pbl(
            zmin=300.0, zmax=3000.0, under_clouds=True, min_snr=1.5, verbose=verbose
        )

        # retrievals
        # inversion method selection
        # 1. default: forward, 50sr
        inversion_method = "forward"
        apriori = {"lr": 50, "mec": False, "use_cfg": False}
        # 2. if exist, overwrite with CFG["parameters"]
        var_apriori = CFG["parameters"][profiles.data.instrument_type]["inversion"][
            "apriori"
        ]["var"]
        if profiles.data.instrument_type in CFG["parameters"]:
            if (
                "cfg"
                in CFG["parameters"][profiles.data.instrument_type]["inversion"][
                    "apriori"
                ]
            ):
                cfg_path = CFG["parameters"][profiles.data.instrument_type][
                    "inversion"
                ]["apriori"]["cfg"]
                # open config
                f = open(Path(Path(__file__).parent, "..", "..", cfg_path))
                aer_ifs = json.load(f)
                f.close()
                station_id = f'{profiles._data.attrs["wigos_station_id"]}-{profiles._data.attrs["instrument_id"]}'
                if station_id in aer_ifs:
                    apriori = {
                        var_apriori: aer_ifs[station_id][var_apriori],
                        "use_cfg": True,
                        "cfg": {
                            "data": aer_ifs[station_id]["data"],
                            "use_default": "False",
                            "attributes": aer_ifs["attributes"],
                            "path": cfg_path,
                        },
                    }
                    if aer_ifs[station_id]["mec"]:
                        apriori["mec"] = aer_ifs[station_id]["mec"]
                else:
                    apriori = {
                        var_apriori: aer_ifs["attributes"]["default"][var_apriori],
                        "use_cfg": True,
                        "cfg": {
                            "use_default": "True",
                            "attributes": aer_ifs["attributes"],
                            "path": cfg_path,
                        },
                    }
                    if aer_ifs["attributes"]["default"]["mec"]:
                        apriori["mec"] = aer_ifs["attributes"]["default"]["mec"]

            if "inversion" in CFG["parameters"][profiles.data.instrument_type]:
                if (
                    "method"
                    in CFG["parameters"][profiles.data.instrument_type]["inversion"]
                ):
                    inversion_method = CFG["parameters"][profiles.data.instrument_type][
                        "inversion"
                    ]["method"]

        profiles.inversion(
            zmin=4000.0,
            zmax=6000.0,
            remove_outliers=True,
            method=inversion_method,
            apriori=apriori,
            verbose=verbose,
        )
        profiles.write(base_dir)
